/*
 * Copyright (C) 2012 Tokyo Institute of Technology 
 *
 * 
 * This file is part of MEGADOCK.
 * MEGADOCK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MEGADOCK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MEGADOCK.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : Control
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "control.h"

//============================================================================//
void Control::initialize(int argc,char *argv[])
//============================================================================//
{
  int			ngrid;
  vector<int>		ngrid_table;
  time_t 		ct1, ct2;

  time(&ct1);
 
  // Parameter
  _parameter = new Parameter(_parallel);
  _parameter->initialize(argc,argv);

  // Receptor
  _receptor = new Receptor(_parameter->_RecPDB_file);
  _receptor->initialize(_parameter);

  // Ligand
  _ligand = new Ligand(_parameter->_LigPDB_file);
  _ligand->initialize(_parameter);

  if( !_parameter->_Num_fft_flag ) {
    gridtable(ngrid,ngrid_table);
    autogridr(ngrid,ngrid_table);
    autogridl(ngrid,ngrid_table);
  }
  else {
    checkgridr();
    checkgridl();
  }

  // Docking
  _docking = new Docking(_cputime,_parallel,_parameter,_receptor,_ligand);
  _docking->initialize();

  time(&ct2);
  _cputime->control_initialize += difftime(ct2,ct1);

  return;
}

//============================================================================//
void Control::gridtable(int &ngrid,vector<int> &ngrid_table)
//============================================================================//
{
  int grid_table[] = {
    2,
    3,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    14,
    15,
    16,
    18,
    20,
    21,
    22,
    24,
    25,
    27,
    28,
    30,
    32,
    33,
    35,
    36,
    40,
    42,
    44,
    45,
    48,
    49,
    50,
    54,
    55,
    56,
    60,
    63,
    64,
    66,
    70,
    72,
    75,
    77,
    80,
    81,
    84,
    88,
    90,
    96,
    98,
    99,
    100,
    105,
    108,
    110,
    112,
    120,
    121,
    125,
    126,
    128,
    132,
    135,
    140,
    144,
    147,
    150,
    154,
    160,
    162,
    165,
    168,
    175,
    176,
    180,
    189,
    192,
    196,
    198,
    200,
    210,
    216,
    220,
    224,
    225,
    231,
    240,
    242,
    243,
    245,
    250,
    252,
    256,
    264,
    270,
    275,
    280,
    288,
    294,
    297,
    300,
    308,
    315,
    320,
    324,
    330,
    336,
    343,
    350,
    352,
    360,
    363,
    375,
    378,
    384,
    385,
    392,
    396,
    400,
    405,
    420,
    432,
    440,
    441,
    448,
    450,
    462,
    480,
    484,
    486,
    490,
    495,
    500,
    504,
    512,
    525,
    528,
    539,
    540,
    550,
    560,
    567,
    576,
    588,
    594,
    600,
    605,
    616,
    625,
    630,
    640,
    648,
    660,
    672,
    675,
    686,
    693,
    700,
    704,
    720,
    726,
    729,
    735,
    750,
    756,
    768,
    770,
    784,
    792,
    800,
    810,
    825,
    840,
    847,
    864,
    875,
    880,
    882,
    891,
    896,
    900,
    924,
    945,
    960,
    968,
    972,
    980,
    990,
    1000,
    1008,
    1024
  };

  ngrid = sizeof(grid_table)/sizeof(grid_table[0]);
  ngrid_table.resize(ngrid);
  if( ngrid_table.size() < ngrid ) {
    cerr << "[ERROR] Out of memory. Number of FFT table lists = ("
	 << ngrid << ") for (ngrid_table) in control.cpp!!\n";
    exit(1);
  }

  for( int i = 0 ; i < ngrid ; i++ ) {
    ngrid_table[i] = grid_table[i];
  }

  return;
}

//============================================================================//
void Control::autogridr(const int &ngrid,vector<int> &ngrid_table)
//============================================================================//
{
  int		num_grid = 1;
  float		size, size_rec = 0.0;

  for( int i = 0 ; i < 3 ; i++ ) {
    size = _receptor->edge(i,1) - _receptor->edge(i,0);

    if( size > size_rec ) {
      size_rec = size;
    }
  }

  cout << "\nReceptor = " << _receptor->input_file() << endl;
  cout << "Receptor max size = " << size_rec << endl;

  size_rec += 2.0 * _parameter->_Grid_space_rec;
  cout << "Required voxel size = " << size_rec << endl;

  num_grid = 1 + int(size_rec / _parameter->_Grid_size);

  for( int i = 0 ; i < ngrid ; i++ ) {
    if( ngrid_table[i] >= num_grid ) {
      num_grid = ngrid_table[i];
      break;
    }
  }

  _receptor->num_grid(num_grid);

  cout << "Number of grid = " << num_grid << endl;
  cout << "FFT N = " << num_grid*2 << endl;

  return;
}

//============================================================================//
void Control::autogridl(const int &ngrid,vector<int> &ngrid_table)
//============================================================================//
{
  int		num_grid = 1;
  float		size_lig = 0.0;
  float		x1, y1, z1, x2, y2, z2, d2;
  const int	na	= _ligand->num_atoms();

  for( int i = 0 ; i < na-1 ; i++ ) {
    x1 = _ligand->coordinate(i,0);
    y1 = _ligand->coordinate(i,1);
    z1 = _ligand->coordinate(i,2);

    for( int j = i+1 ; j < na ; j++ ) {
      x2 = _ligand->coordinate(j,0);
      y2 = _ligand->coordinate(j,1);
      z2 = _ligand->coordinate(j,2);

      d2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);

      if( d2 > size_lig ) {
	size_lig = d2;
      }
    }
  }

  size_lig = sqrt(size_lig);

  cout << "\nLigand = " << _ligand->input_file() << endl;
  cout << "Ligand max size = " << size_lig << endl;

  size_lig += 2.0 * _parameter->_Grid_space_lig;

  cout << "Required voxel size = " << size_lig << endl;

  num_grid = 1 + int(size_lig / _parameter->_Grid_size);

  for( int i = 0 ; i < ngrid ; i++ ) {
    if( ngrid_table[i] >= num_grid ) {
      num_grid = ngrid_table[i];
      break;
    }
  }

  _ligand->num_grid(num_grid);

  cout << "Number of grid = " << num_grid << endl;
  cout << "FFT N = " << num_grid*2 << endl;

  return;
}

//============================================================================//
void Control::checkgridr()
//============================================================================//
{
  float		size, size_rec	= 0.0;
  const int	num_grid	= _parameter->_Num_grid;
  const float	search_length	= _parameter->_Grid_size * num_grid;

  for( int i = 0 ; i < 3 ; i++ ) {
    size = _receptor->edge(i,1) - _receptor->edge(i,0);

    if( size > size_rec ) {
      size_rec = size;
    }
  }

  cout << "\nReceptor max size = " << size_rec << endl;

  size_rec += 2.0*_parameter->_Grid_space_rec;

  cout << "Required voxel size = " << size_rec << endl;

  if( size_rec > search_length ) {
    cerr << "[ERROR] Receptor data is too big!!\n";
    exit(1);
  }

  _receptor->num_grid(num_grid);

  cout << "\n(Receptor)\n";
  cout << "Number of grid = " << num_grid << endl;
  cout << "FFT N = " << num_grid*2 << endl;
  cout << "Grid size = " << _parameter->_Grid_size << endl;

  return;
}

//============================================================================//
void Control::checkgridl()
//============================================================================//
{
  float		size_lig = 0.0;
  float		x1, y1, z1, x2, y2, z2, d2;
  const int	na	= _ligand->num_atoms();
  const int	num_grid	= _parameter->_Num_grid;
  const float	search_length	= _parameter->_Grid_size * num_grid;

  for( int i = 0 ; i < na-1 ; i++ ) {
    x1 = _ligand->coordinate(i,0);
    y1 = _ligand->coordinate(i,1);
    z1 = _ligand->coordinate(i,2);

    for( int j = i+1 ; j < na ; j++ ) {
      x2 = _ligand->coordinate(j,0);
      y2 = _ligand->coordinate(j,1);
      z2 = _ligand->coordinate(j,2);

      d2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);

      if( d2 > size_lig ) {
	size_lig = d2;
      }
    }
  }

  size_lig = sqrt(size_lig);
  cout << "\nLigand max size = " << size_lig << endl;

  size_lig += 2.0*_parameter->_Grid_space_lig;
  cout << "Required voxel size = " << size_lig << endl;

  if( size_lig > search_length ) {
    cerr << "[ERROR] Ligand data is too big!!\n";
    exit(1);
  }

  _ligand->num_grid(num_grid);

  cout << "\n(Ligand)\n";
  cout << "Number of grid = " << num_grid << endl;
  cout << "FFT N = " << num_grid*2 << endl;
  cout << "Grid size = " << _parameter->_Grid_size << endl;

  return;
}

//============================================================================//
void Control::execute()
//============================================================================//
{
  time_t 	ct1, ct2;

  cout << "\n---------- Start docking calculations" << endl;

  time(&ct1);
  _docking->rec_init();
  time(&ct2);
  _cputime->receptor_process += difftime(ct2,ct1);

  time(&ct1);
  _docking->dockz();
  time(&ct2);
  _cputime->docking_dock += difftime(ct2,ct1);

  time(&ct1);
  _docking->output();
  time(&ct2);
  _cputime->docking_output += difftime(ct2,ct1);

  delete _docking;
  delete _ligand;
  delete _receptor;
  delete _parameter;

  return;
}
