/*
 * Copyright (C) 2012 Tokyo Institute of Technology 
 *
 * 
 * This file is part of MEGADOCK.
 * MEGADOCK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MEGADOCK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MEGADOCK.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : FFTProcess
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "fft_process.h"

//============================================================================//
bool ascending_struct(const SortScore &s1,const SortScore &s2)	// for sort
//============================================================================//
{
  return s1.score > s2.score;
}

//============================================================================//
void FFTProcess::alloc_array(const int &num_fft)
//============================================================================//
{
  _Num_fft = num_fft;
  const int 	nf3 	= _Num_fft * _Num_fft * _Num_fft;
  const int 	ns 	= _parameter->_Num_sort;
  const int 	nz 	= _parameter->_NZangle;
  const int 	no 	= _parameter->_Num_output;
  const int	nproc2	= _parallel->nproc2();
  int	nt;

  nt = nz * ns;
  if( no > nt ) nt = no;

  alloc_fft();

  _Select.resize(nproc2);
  for( int i = 0 ; i < nproc2 ; i++ ) {
    _Select[i].resize(ns);
  }

  _Top.resize(nt);

  //---------- memory allocation for _Rotation
  _Rotation = new int[nproc2];

  //---------- memory allocation for _FFT_rec_r
  _FFT_rec_r = new float[nf3];
  if( !_FFT_rec_r ) {
    cerr << "[ERROR] Out of memory. Number of listed receptors = ("
	 << nf3 << ") for (_FFT_rec_r) in fft_process.cpp!!\n";
    exit(1);
  }

  //---------- memory allocation for _FFT_rec_i
  _FFT_rec_i = new float[nf3];
  if( !_FFT_rec_i ) {
    cerr << "[ERROR] Out of memory. Number of listed receptors = ("
	 << nf3 << ") for (_FFT_rec_i) in fft_process.cpp!!\n";
    exit(1);
  }

  //---------- memory allocation for _FFT_lig_r
  _FFT_lig_r = new float[nproc2*nf3];
  if( !_FFT_lig_r ) {
      cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
	   << nf3*nproc2 << ") for (_FFT_lig_r) in fft_process.cpp!!\n";
      exit(1);
  }

  //---------- memory allocation for _FFT_lig_i
  _FFT_lig_i = new float[nproc2*nf3];
  if( !_FFT_lig_i ) {
      cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
	   << nf3*nproc2 << ") for (_FFT_lig_i) in fft_process.cpp!!\n";
      exit(1);
  }

  return;
}

//============================================================================//
void FFTProcess::alloc_fft()
//============================================================================//
{
  const int	nf3 = _Num_fft * _Num_fft * _Num_fft;
  const int	nproc2	= _parallel->nproc2();

  _FFTWin  = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*nf3*nproc2);
  _FFTWout = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*nf3*nproc2);

  return;
}

//============================================================================//
void FFTProcess::receptor_fft(float *grid_r,float *grid_i)
//============================================================================//
{
  const int	num_grid= _Num_fft / 2;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  const int	ndata	= ( _Num_fft - num_grid ) / 2;
  const float	theta	= -2.0 * PI / _Num_fft;

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[i][0] = 0.0;
    _FFTWin[i][1] = 0.0;
  }

  for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
    const int ic = _Num_fft*_Num_fft*(i+ndata);

    for( int j = 0 ; j < num_grid ; j++ ) {
      const int jc = ic + _Num_fft*(j+ndata);

      for( int k = 0 ; k < num_grid ; k++ ) {
	_FFTWin[jc+k+ndata][0] = grid_r[m  ];
	_FFTWin[jc+k+ndata][1] = grid_i[m++];
      }
    }
  }

  fft3d(theta,0);

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFT_rec_r[i] = _FFTWout[i][0];
    _FFT_rec_i[i] = _FFTWout[i][1];
  }

  return;
}

//============================================================================//
void FFTProcess::ligand_fft(float *grid_r,float *grid_i,int myid2)
//============================================================================//
{
  const int 	num_grid = _Num_fft / 2;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  const int	ndata	= ( _Num_fft - num_grid ) / 2;
  const float	theta	= -2.0 * PI / _Num_fft;
  const int	np = nf3 * myid2;

  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[np+i][0] = 0.0;
    _FFTWin[np+i][1] = 0.0;
  }

  for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
    const int ic = _Num_fft*_Num_fft*(i+ndata);

    for( int j = 0 ; j < num_grid ; j++ ) {
      int jc = ic + _Num_fft*(j+ndata);

      for( int k = 0 ; k < num_grid ; k++ ) {
	_FFTWin[np+jc+k+ndata][0] = grid_r[m  ];
	_FFTWin[np+jc+k+ndata][1] = grid_i[m++];
      }
    }
  }

  fft3d(theta,myid2);

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFT_lig_r[np+i] = _FFTWout[np+i][0];
    _FFT_lig_i[np+i] = _FFTWout[np+i][1];
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_ligand_fft += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::fft_score(int myid2)
//============================================================================//
{
  const int	nf3 	= _Num_fft * _Num_fft * _Num_fft;
  const float	theta 	= 2.0 * PI / _Num_fft;
  const int	np	= nf3 * myid2;
  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[np+i][0] = _FFT_rec_r[i]*_FFT_lig_r[np+i]
      + _FFT_rec_i[i]*_FFT_lig_i[np+i];
    _FFTWin[np+i][1] = _FFT_rec_r[i]*_FFT_lig_i[np+i]
      - _FFT_rec_i[i]*_FFT_lig_r[np+i];
  }

  fft3d(theta,myid2);

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_fft_inverse += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::fft3d(const float &theta,int myid2)
//============================================================================//
{
  const int	n   = _Num_fft;
  const int	np = n*n*n * myid2;
  const int	nproc2 = _parallel->nproc2();
  fftw_plan	plan;

#pragma omp critical
  {
    if( theta < 0.0 ) {
      plan=fftw_plan_dft_3d(n,n,n,&_FFTWin[np],&_FFTWout[np],
			     FFTW_FORWARD,FFTW_ESTIMATE); 
    }
    else {
      plan=fftw_plan_dft_3d(n,n,n,&_FFTWin[np],&_FFTWout[np],
			     FFTW_BACKWARD,FFTW_ESTIMATE); 
    }
  }

  fftw_execute(plan); 

#pragma omp critical
  {
    fftw_destroy_plan(plan);
  }

  return;
}

//============================================================================//
void FFTProcess::score_sort(int myid2)
//============================================================================//
{
  const int	ns	= _parameter->_Num_sort;
  const int	nf2	= _Num_fft * _Num_fft;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  const int	np	= nf3 * myid2;
  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < ns ; i++ ) {
    _Select[myid2][i].score = 0.0;
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    const float raw = _FFTWout[np+i][0] / nf3;

    for( int j = 0 ; j < ns ; j++ ) {
      if( raw > _Select[myid2][j].score ) {
	for( int k = ns-1 ; k > j ; k-- ) {
	  _Select[myid2][k] = _Select[myid2][k-1];
	}
	_Select[myid2][j].score	   = raw;
	_Select[myid2][j].index[1] = i / nf2;
	_Select[myid2][j].index[2] = (i / _Num_fft) % _Num_fft;
	_Select[myid2][j].index[3] = i % _Num_fft;
	break;
      }
    }
  }

  for( int i = 0 ; i < ns ; i++ ) {
    _Select[myid2][i].index[0] = _Rotation[myid2];
  }

  for( int i = 0 ; i < ns ; i++ ) {
    _Top[_Rotation[myid2]*ns+i] = _Select[myid2][i];
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_score_sort += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::sort_index(float *fwork,int *iwork)
//============================================================================//
{
  const int 	no 	= _parameter->_Num_output;
  const int 	nt 	= (_Top.size() < no) ? _Top.size() : no;

  partial_sort(_Top.begin(),_Top.begin()+nt,_Top.end(),
	       ascending_struct);

  return;
}

//============================================================================//
void FFTProcess::top_score_clean()
//============================================================================//
{
  const int 	ns 	= _parameter->_Num_sort;
  const int 	nz 	= _parameter->_NZangle;
  const int 	no 	= _parameter->_Num_output;
  int	nt;

  nt = nz * ns;
  if( no > nt ) nt = no;

  for( int j = 0 ; j < nt ; j++ ) {
    _Top[j].score = 0.0;
  }

  return;
}
