/*
 * Copyright (C) 2012 Tokyo Institute of Technology 
 *
 * 
 * This file is part of MEGADOCK.
 * MEGADOCK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MEGADOCK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MEGADOCK.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : Docking
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "docking.h"

//============================================================================//
void Docking::initialize()
//============================================================================//
{
  _Num_grid = max(_receptor->num_grid(),_ligand->num_grid());

  // FFTProcess
  _fft_process	= new FFTProcess(_cputime,_parallel,_parameter);

  maxsize_voxel();

  return;
}

//============================================================================//
void Docking::maxsize_voxel()
//============================================================================//
{
  const int 	num_fft = _Num_grid * 2;
  const int 	natom = max(_receptor->num_atoms(),_ligand->num_atoms());
  const int	nag = natom * _Num_grid;
  const int	ng3 = _Num_grid * _Num_grid * _Num_grid;
  const int 	nf3 = num_fft * num_fft * num_fft;

  const int 	ns 	= _parameter->_Num_sort;
  const int 	nz 	= _parameter->_NZangle;
  const int	nproc2	= _parallel->nproc2();
  float		fmem;
  float		imem;
    
  fmem = nproc2*(ng3*3+natom*3+nag*3)
    +nproc2*2*nf3+2*nf3
    +nproc2*4*nf3
    +nz*ns;
  imem = nproc2*(ng3*2+natom*4)
    +4*nz*ns;

  double	mb = (fmem*4+imem*2)/1024/1024;
  mb += 10.0;		// margine
  printf("Memory requirement (/node)  = %.1f MB\n",mb);

  _fft_process->alloc_array(num_fft);
  alloc_array(natom,nag,ng3);

  return;
}

//============================================================================//
void Docking::alloc_array(const int &natom,const int &nag,const int &ng3)
//============================================================================//
{
  const int	nproc2	  = _parallel->nproc2();

  _Grid_coord = new float[_Num_grid];
  if( !_Grid_coord ) {
    cerr << "[ERROR] Out of memory. Number of voxels for 1 axis = ("
	 << _Num_grid
	 << ") for (_Grid_coord) in docking.cpp!!\n";
    exit(1);
  }

  _Mol_coord = new float[nproc2*natom*3];
  if( !_Mol_coord ) {
	  cerr << "[ERROR] Out of memory. Number of atoms * 3 * number of threads = )"
		  << natom*3*nproc2
	   << ") for (_Mol_coord) in docking.cpp!!\n";
      exit(1);
   }

  _Memfw = ng3*3+natom*3+nag*3;
  _Fwork = new float[_Memfw*nproc2];
  if( !_Fwork ) {
    cerr << "[ERROR] Out of memory. Number of float arrays = ("
	 << _Memfw*nproc2
	 << ") for (_Fwork) in docking.cpp!!\n";
    exit(1);
  }

  _Memiw = ng3*2+natom*4;
  _Iwork = new int[_Memiw*nproc2];
  if( !_Iwork ) {
    cerr << "[ERROR] Out of memory. Number of int arrays = ("
	 << _Memiw*nproc2
	 << ") for (_Iwork) in docking.cpp!!\n";
    exit(1);
  }

  return;
}

//============================================================================//
void Docking::rec_init()
//============================================================================//
{
  const int	ng3 	= _Num_grid * _Num_grid * _Num_grid;
  float	search_length	= _parameter->_Grid_size * _Num_grid;

  for( int i = 0 ; i < _Num_grid ; i++ ) {
    _Grid_coord[i] = search_length * ( -0.5 + ( 0.5 + i )/_Num_grid );
  }

  _receptor->shift_center(_Mol_coord);
  create_voxel(_receptor,0);
  _fft_process->receptor_fft(_Fwork,&_Fwork[ng3]);

  return;
}

//============================================================================//
void Docking::create_voxel(Protein *rprotein,int myid2)
//============================================================================//
{
  time_t	ct1, ct2;
  const int	np = rprotein->num_atoms() * 3 * myid2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  if( _parameter->_Score_func == 2 ) {
    rprotein->rpsc(_Num_grid,_Grid_coord,&_Mol_coord[np],&_Iwork[_Memiw*myid2],
		  &_Fwork[_Memfw*myid2]);
  }
  else if( _parameter->_Score_func == 1 ) {	// CHARMM19 electrostatic
    const float	beta = -2000.0;			// MEGADOCK parameter
    rprotein->rpsc(_Num_grid,_Grid_coord,&_Mol_coord[np],&_Iwork[_Memiw*myid2],
		  &_Fwork[_Memfw*myid2]);
    rprotein->electro(beta,_parameter->_Elec_ratio,_Num_grid,_Grid_coord,
		     &_Iwork[_Memiw*myid2],&_Fwork[_Memfw*myid2]);
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->protein_lig_voxelization += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void Docking::dockz()
//============================================================================//
{
  time_t 	ct1, ct2;
  float		theta[3];

  const int	ng3 	  = _Num_grid * _Num_grid * _Num_grid;
  const int 	nc	  = _parameter->_NZangle / 10;

  _fft_process->top_score_clean();

  cout << "\nLigand = " << _parameter->_LigPDB_file << endl;
  cout << "Target receptors:" << endl;;
  cout << " " << _parameter->_RecPDB_file << endl;
  cout << endl;
  cout.flush();

  int 	j, myid2;

#pragma omp parallel private(myid2,j,theta)
  {
#ifdef _OPENMP
    myid2 = omp_get_thread_num();
#else
    myid2 = 0;
#endif

#pragma omp for
    for( int i = 0 ; i < _parameter->_NZangle ; i++ ) {
      if( !((i+1)%nc) ) {
	cout << "   >Ligand rotation = " << i+1 << " / "
	     << _parameter->_NZangle << " (" << myid2 << ")\n";
	cout.flush();
      }

      _fft_process->rotation_index(myid2,i);
      
      for( j = 0 ; j < 3 ; j++ ) {
	theta[j] = _parameter->_Zangle[3*i+j];
      }

      ligand_rotationz(theta,myid2);
      create_voxel(_ligand,myid2);
      _fft_process->ligand_fft(&_Fwork[_Memfw*myid2],&_Fwork[_Memfw*myid2+ng3],
      			       myid2);

      _fft_process->fft_score(myid2);
      _fft_process->score_sort(myid2);
    }
  }

  time(&ct1);
  _fft_process->sort_index(_Fwork,_Iwork);
  time(&ct2);
  _cputime->fftprocess_sort_index += difftime(ct2,ct1);

  return;
}

//============================================================================//
void Docking::ligand_rotationz(float *theta,int myid2)
//============================================================================//
{
  const int	nl = _ligand->num_atoms();
  const int	np = nl * 3 * myid2;
  float		x, y, z;
  float		r11, r21, r31, r12, r22, r32, r13, r23, r33;

  r11 = cos(theta[0])*cos(theta[2])  -  sin(theta[0])*cos(theta[1])*sin(theta[2]);
  r21 = sin(theta[0])*cos(theta[2])  +  cos(theta[0])*cos(theta[1])*sin(theta[2]);
  r31 = sin(theta[1])*sin(theta[2]);

  r12 = -cos(theta[0])*sin(theta[2])  -  sin(theta[0])*cos(theta[1])*cos(theta[2]);
  r22 = -sin(theta[0])*sin(theta[2])  +  cos(theta[0])*cos(theta[1])*cos(theta[2]);
  r32 = sin(theta[1])*cos(theta[2]);

  r13 = sin(theta[0])*sin(theta[1]);
  r23 = -cos(theta[0])*sin(theta[1]);
  r33 = cos(theta[1]);

  for( int i = 0 ; i < nl ; i++ ) {
    x = _ligand->coordinate(i,0) - _ligand->_Center[0];
    y = _ligand->coordinate(i,1) - _ligand->_Center[1];
    z = _ligand->coordinate(i,2) - _ligand->_Center[2];
    _Mol_coord[ np+3*i  ] = r11 * x + r12 * y + r13 * z;
    _Mol_coord[ np+3*i+1] = r21 * x + r22 * y + r23 * z;
    _Mol_coord[ np+3*i+2] = r31 * x + r32 * y + r33 * z;
  }

  return;
}

//============================================================================//
void Docking::output()
//============================================================================//
{
  char		buff[256];
  float		coord[3], coordl[3], theta[3];

  const int	num_fft	= _Num_grid * 2;
  const float	bin 	= _parameter->_Grid_size;
  const int	no 	= _parameter->_Num_output;

  for( int i = 0 ; i < 3 ; i++ ) {
    coordl[i] = _ligand->center(i);
  }

  ofstream Output(_parameter->_RLOut_file.c_str());
  if( !Output ) {
    cerr << "[ERROR] Output file [" << _parameter->_RLOut_file
	 << "] not open!!" << endl;
    exit(1);
  }

  sprintf(buff,"%d\t%.2f",num_fft,bin);
  Output << buff << endl;

  for( int j = 0 ; j < 3 ; j++ ) {
    theta[j] = _receptor->angle(j);
  }

  sprintf(buff,"\t%f\t%f\t%f",theta[0],theta[1],theta[2]);
  Output << buff << endl;

  for( int j = 0 ; j < 3 ; j++ ) {
    coord[j] = _receptor->center(j);
  }

  sprintf(buff,"\t%f\t%f\t%f",coord[0],coord[1],coord[2]);

  Output << _parameter->_RecPDB_file << buff << endl;

  sprintf(buff,"\t%f\t%f\t%f",coordl[0],coordl[1],coordl[2]);
  Output << _parameter->_LigPDB_file << buff << endl;

  for( int j = 0 ; j < no ; j++ ) {
    const int n = _fft_process->top_index(j,0);

    for( int k = 0 ; k < 3 ; k++ ) {
      theta[k] = _parameter->_Zangle[3*n+k];

      if( theta[k] >= PI ) {
	theta[k] -= PI*2.0;
      }
    }

    sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t%.2f",theta[0],theta[1],theta[2],
	    _fft_process->top_index(j,1),_fft_process->top_index(j,2),
	    _fft_process->top_index(j,3),_fft_process->top_score(j));
    Output << buff << endl;
  }

  Output.close();

  return;
}
